@@ -24,10 +24,11 @@
 #include <algorithm>
 #include <iostream>
 #include <iterator>
+#include <optional>
 #include <type_traits>
 
-#include "../config.hpp"
 #include "../common.hpp"
+#include "../config.hpp"
 #include "../detail/temp_storage.hpp"
 #include "../detail/various.hpp"
 #include "../functional.hpp"
@@ -51,9 +52,9 @@ BEGIN_ROCPRIM_NAMESPACE
 namespace detail
 {
 
-template<select_method SelectMethod,
+template<class Config,
+         select_method SelectMethod,
          bool          OnlySelected,
-         class Config,
          class KeyIterator,
          class ValueIterator,
          class FlagIterator,
@@ -63,85 +64,120 @@ template<select_method SelectMethod,
          class OffsetLookbackScanState,
          class BlockIdWrapper,
          class... UnaryPredicates>
-ROCPRIM_KERNEL ROCPRIM_LAUNCH_BOUNDS(device_params<Config>().kernel_config.block_size) void
-    partition_kernel(KeyIterator                        keys_input,
-                     ValueIterator                      values_input,
-                     FlagIterator                       flags,
-                     OutputKeyIterator                  keys_output,
-                     OutputValueIterator                values_output,
-                     size_t*                            selected_count,
-                     size_t*                            prev_selected_count,
-                     size_t                             prev_processed,
-                     const size_t                       total_size,
-                     InequalityOp                       inequality_op,
-                     OffsetLookbackScanState            offset_scan_state,
-                     const unsigned int                 number_of_blocks,
-                     detail::vsmem_t                    vsmem,
-                     BlockIdWrapper                     block_id,
-                     UnaryPredicates... predicates)
+inline hipError_t launch_partition(detail::target_arch     arch,
+                                   KeyIterator             keys_input,
+                                   ValueIterator           values_input,
+                                   FlagIterator            flags,
+                                   OutputKeyIterator       keys_output,
+                                   OutputValueIterator     values_output,
+                                   size_t*                 selected_count,
+                                   size_t*                 prev_selected_count,
+                                   size_t                  prev_processed,
+                                   const size_t            total_size,
+                                   InequalityOp            inequality_op,
+                                   OffsetLookbackScanState offset_scan_state,
+                                   const unsigned int      number_of_blocks,
+                                   detail::vsmem_t         vsmem,
+                                   BlockIdWrapper          block_id,
+                                   dim3                    grid,
+                                   dim3                    block,
+                                   size_t                  shmem,
+                                   hipStream_t             stream,
+                                   UnaryPredicates... predicates)
 {
-    using offset_type = typename OffsetLookbackScanState::value_type;
-    using key_type    = typename std::iterator_traits<KeyIterator>::value_type;
-    using value_type  = typename std::iterator_traits<ValueIterator>::value_type;
-    using flag_type =
-        typename std::conditional<SelectMethod == select_method::predicated_flag,
-                                  typename std::iterator_traits<FlagIterator>::value_type,
-                                  bool>::type;
+    auto kernel = [=](auto arch_config) mutable
+    {
+        using offset_type = typename OffsetLookbackScanState::value_type;
+        using key_type    = typename std::iterator_traits<KeyIterator>::value_type;
+        using value_type  = typename std::iterator_traits<ValueIterator>::value_type;
+        using flag_type =
+            typename std::conditional<SelectMethod == select_method::predicated_flag,
+                                      typename std::iterator_traits<FlagIterator>::value_type,
+                                      bool>::type;
 
-    using partition_kernel_impl_t = partition_kernel_impl_<SelectMethod,
-                                                           OnlySelected,
-                                                           Config,
-                                                           key_type,
-                                                           value_type,
-                                                           flag_type,
-                                                           offset_type,
-                                                           BlockIdWrapper>;
+        using partition_kernel_impl_t = partition_kernel_impl_<decltype(arch_config),
+                                                               SelectMethod,
+                                                               OnlySelected,
+                                                               key_type,
+                                                               value_type,
+                                                               flag_type,
+                                                               offset_type,
+                                                               BlockIdWrapper>;
 
-    using VSmemHelperT = detail::vsmem_helper_impl<partition_kernel_impl_t>;
-    ROCPRIM_SHARED_MEMORY typename VSmemHelperT::static_temp_storage_t static_temp_storage;
-    // Get temporary storage
-    typename partition_kernel_impl_t::storage_type& storage
-        = VSmemHelperT::get_temp_storage(static_temp_storage, vsmem);
+        using VSmemHelperT = detail::vsmem_helper_impl<partition_kernel_impl_t>;
+        ROCPRIM_SHARED_MEMORY typename VSmemHelperT::static_temp_storage_t static_temp_storage;
+        // Get temporary storage
+        typename partition_kernel_impl_t::storage_type& storage
+            = VSmemHelperT::get_temp_storage(static_temp_storage, vsmem);
 
-    partition_kernel_impl_t().partition(keys_input,
-                                        values_input,
-                                        flags,
-                                        keys_output,
-                                        values_output,
-                                        selected_count,
-                                        prev_selected_count,
-                                        prev_processed,
-                                        total_size,
-                                        inequality_op,
-                                        offset_scan_state,
-                                        number_of_blocks,
-                                        block_id,
-                                        storage,
-                                        predicates...);
+        partition_kernel_impl_t().partition(keys_input,
+                                            values_input,
+                                            flags,
+                                            keys_output,
+                                            values_output,
+                                            selected_count,
+                                            prev_selected_count,
+                                            prev_processed,
+                                            total_size,
+                                            inequality_op,
+                                            offset_scan_state,
+                                            number_of_blocks,
+                                            block_id,
+                                            storage,
+                                            predicates...);
+    };
+
+    return execute_launch_plan<Config>(arch, kernel, grid, block, shmem, stream);
 }
 
-template<select_method SelectMethod,
+template<class Config,
+         select_method SelectMethod,
          bool          OnlySelected,
-         class Config,
          class Key,
          class Value,
          class FlagType,
          class OffsetLookbackScanState,
          class BlockIdWrapper>
-inline size_t get_partition_vsmem_size_per_block()
+inline size_t get_partition_vsmem_size_per_block(detail::target_arch arch)
 {
-    using offset_type             = typename OffsetLookbackScanState::value_type;
-    using partition_kernel_impl_t = partition_kernel_impl_<SelectMethod,
-                                                           OnlySelected,
-                                                           Config,
-                                                           Key,
-                                                           Value,
-                                                           FlagType,
-                                                           offset_type,
-                                                           BlockIdWrapper>;
+    using offset_type = typename OffsetLookbackScanState::value_type;
+    std::optional<size_t> vsmem_per_block;
+    for_each_arch(
+        [&](auto arch_tag)
+        {
+            constexpr target_arch Arch = decltype(arch_tag)::value;
+            if(Arch != arch || vsmem_per_block)
+                return;
 
-    using PartitionVSmemHelperT = detail::vsmem_helper_impl<partition_kernel_impl_t>;
-    return PartitionVSmemHelperT::vsmem_per_block;
+            using ArchConfig               = typename Config::template architecture_config<Arch>;
+            using partition_kernel_impl_t  = partition_kernel_impl_<ArchConfig,
+                                                                   SelectMethod,
+                                                                   OnlySelected,
+                                                                   Key,
+                                                                   Value,
+                                                                   FlagType,
+                                                                   offset_type,
+                                                                   BlockIdWrapper>;
+            using partition_vsmem_helper_t = detail::vsmem_helper_impl<partition_kernel_impl_t>;
+
+            vsmem_per_block = partition_vsmem_helper_t::vsmem_per_block;
+        });
+    if(!vsmem_per_block)
+    {
+        using ArchConfig = typename Config::template architecture_config<target_arch::unknown>;
+        using partition_kernel_impl_t  = partition_kernel_impl_<ArchConfig,
+                                                               SelectMethod,
+                                                               OnlySelected,
+                                                               Key,
+                                                               Value,
+                                                               FlagType,
+                                                               offset_type,
+                                                               BlockIdWrapper>;
+        using partition_vsmem_helper_t = detail::vsmem_helper_impl<partition_kernel_impl_t>;
+
+        vsmem_per_block = partition_vsmem_helper_t::vsmem_per_block;
+    }
+    return vsmem_per_block.value();
 }
 
 template<partition_subalgo SubAlgo,
@@ -171,8 +207,8 @@ inline hipError_t partition_impl(void*                       temporary_storage,
                                  UnaryPredicates... predicates)
 {
     using offset_type = OffsetT;
-    using key_type = typename std::iterator_traits<KeyIterator>::value_type;
-    using value_type = typename std::iterator_traits<ValueIterator>::value_type;
+    using key_type    = typename std::iterator_traits<KeyIterator>::value_type;
+    using value_type  = typename std::iterator_traits<ValueIterator>::value_type;
 
     using config = wrapped_partition_config<Config, SubAlgo, key_type, value_type>;
 
@@ -200,16 +236,16 @@ inline hipError_t partition_impl(void*                       temporary_storage,
     {
         return result;
     }
-    const partition_config_params params = dispatch_target_arch<config>(target_arch);
+    const partition_config_params params = dispatch_target_arch<config, false>(target_arch);
 
-    using offset_scan_state_type = detail::lookback_scan_state<offset_type>;
+    using offset_scan_state_type            = detail::lookback_scan_state<offset_type>;
     using offset_scan_state_with_sleep_type = detail::lookback_scan_state<offset_type, true>;
 
     const unsigned int block_size       = params.kernel_config.block_size;
     const unsigned int items_per_thread = params.kernel_config.items_per_thread;
     const auto         items_per_block  = block_size * items_per_thread;
 
-    static constexpr bool is_three_way = sizeof...(UnaryPredicates) == 2;
+    static constexpr bool         is_three_way        = sizeof...(UnaryPredicates) == 2;
     static constexpr const size_t selected_count_size = is_three_way ? 2 : 1;
 
     const size_t size_limit = params.kernel_config.size_limit;
@@ -222,9 +258,9 @@ inline hipError_t partition_impl(void*                       temporary_storage,
         = static_cast<unsigned int>(::rocprim::detail::ceiling_div(limited_size, items_per_block));
 
     // Calculate required temporary storage
-    void*                           offset_scan_state_storage;
-    size_t*                         selected_count;
-    size_t*                         prev_selected_count;
+    void*   offset_scan_state_storage;
+    size_t* selected_count;
+    size_t* prev_selected_count;
 
     detail::temp_storage::layout layout{};
     result = offset_scan_state_type::get_temp_storage_layout(number_of_blocks, stream, layout);
@@ -250,25 +286,26 @@ inline hipError_t partition_impl(void*                       temporary_storage,
     if(use_sleep)
     {
         virtual_shared_memory_size
-            = get_partition_vsmem_size_per_block<method,
+            = get_partition_vsmem_size_per_block<config,
+                                                 method,
                                                  write_only_selected,
-                                                 config,
                                                  key_type,
                                                  value_type,
                                                  flag_type,
                                                  offset_scan_state_with_sleep_type,
-                                                 block_id_wrapper_type>();
+                                                 block_id_wrapper_type>(target_arch);
     }
     else
     {
-        virtual_shared_memory_size = get_partition_vsmem_size_per_block<method,
-                                                                        write_only_selected,
-                                                                        config,
-                                                                        key_type,
-                                                                        value_type,
-                                                                        flag_type,
-                                                                        offset_scan_state_type,
-                                                                        block_id_wrapper_type>();
+        virtual_shared_memory_size
+            = get_partition_vsmem_size_per_block<config,
+                                                 method,
+                                                 write_only_selected,
+                                                 key_type,
+                                                 value_type,
+                                                 flag_type,
+                                                 offset_scan_state_type,
+                                                 block_id_wrapper_type>(target_arch);
     }
     virtual_shared_memory_size *= number_of_blocks;
 
@@ -285,7 +322,8 @@ inline hipError_t partition_impl(void*                       temporary_storage,
             // They have the same base type, so there is no padding between the types.
             detail::temp_storage::ptr_aligned_array(&selected_count, selected_count_size),
             detail::temp_storage::ptr_aligned_array(&prev_selected_count, selected_count_size),
-            detail::temp_storage::ptr_aligned_array(&block_id_pool, block_id_wrapper_type::get_storage_size()),
+            detail::temp_storage::ptr_aligned_array(&block_id_pool,
+                                                    block_id_wrapper_type::get_storage_size()),
             // vsmem
             detail::temp_storage::make_partition(&vsmem,
                                                  virtual_shared_memory_size,
@@ -365,7 +403,8 @@ inline hipError_t partition_impl(void*                       temporary_storage,
         const unsigned int current_size
             = static_cast<unsigned int>(std::min<size_t>(size - prev_processed, limited_size));
 
-        const unsigned int current_number_of_blocks = ::rocprim::detail::ceiling_div(current_size, items_per_block);
+        const unsigned int current_number_of_blocks
+            = ::rocprim::detail::ceiling_div(current_size, items_per_block);
 
         if(debug_synchronous)
         {
@@ -389,38 +428,43 @@ inline hipError_t partition_impl(void*                       temporary_storage,
                                                     current_number_of_blocks,
                                                     start);
 
-        if (UsingOrderedBlockId)
+        if(UsingOrderedBlockId)
         {
             result = hipMemsetAsync(block_id_pool, 0, sizeof(unsigned int), stream);
-            if (result != hipSuccess)
+            if(result != hipSuccess)
             {
                 return result;
             }
         }
 
-        if(debug_synchronous) start = std::chrono::steady_clock::now();
+        if(debug_synchronous)
+            start = std::chrono::steady_clock::now();
 
-        with_scan_state(
+        ROCPRIM_RETURN_ON_ERROR(with_scan_state(
             [&](const auto scan_state)
             {
-                partition_kernel<method, write_only_selected, config>
-                    <<<dim3(current_number_of_blocks), dim3(block_size), 0, stream>>>(
-                        keys_input + prev_processed,
-                        values_input + prev_processed,
-                        flags + prev_processed,
-                        keys_output,
-                        values_output,
-                        selected_count,
-                        prev_selected_count,
-                        prev_processed,
-                        size,
-                        inequality_op,
-                        scan_state,
-                        current_number_of_blocks,
-                        detail::vsmem_t{vsmem},
-                        block_id,
-                        predicates...);
-            });
+                return launch_partition<config, method, write_only_selected>(
+                    target_arch,
+                    keys_input + prev_processed,
+                    values_input + prev_processed,
+                    flags + prev_processed,
+                    keys_output,
+                    values_output,
+                    selected_count,
+                    prev_selected_count,
+                    prev_processed,
+                    size,
+                    inequality_op,
+                    scan_state,
+                    current_number_of_blocks,
+                    detail::vsmem_t{vsmem},
+                    block_id,
+                    dim3(current_number_of_blocks),
+                    dim3(block_size),
+                    0,
+                    stream,
+                    predicates...);
+            }));
         ROCPRIM_DETAIL_HIP_SYNC_AND_RETURN_ON_ERROR("partition_kernel", size, start);
 
         std::swap(selected_count, prev_selected_count);
@@ -440,7 +484,7 @@ inline hipError_t partition_impl(void*                       temporary_storage,
     return hipSuccess;
 }
 
-} // end of detail namespace
+} // namespace detail
 
 /// \brief Two-way parallel select primitive for device level using selection predicate.
 ///
@@ -498,7 +542,7 @@ inline hipError_t partition_impl(void*                       temporary_storage,
 /// #include <rocprim/rocprim.hpp>///
 ///
 /// auto predicate =
-///     [] __device__ (int a) -> bool
+///     [] (int a) -> bool
 ///     {
 ///         return (a%2) == 0;
 ///     };
@@ -547,7 +591,7 @@ template<class Config = default_config,
          class RejectedOutputIterator,
          class SelectedCountOutputIterator,
          class Predicate,
-         bool  UsingOrderedBlockId = false>
+         bool UsingOrderedBlockId = false>
 inline hipError_t partition_two_way(void*                       temporary_storage,
                                     size_t&                     storage_size,
                                     InputIterator               input,
@@ -692,7 +736,7 @@ template<class Config = default_config,
          typename SelectedOutputIterator,
          typename RejectedOutputIterator,
          typename SelectedCountOutputIterator,
-         bool     UsingOrderedBlockId = false>
+         bool UsingOrderedBlockId = false>
 inline hipError_t partition_two_way(void*                       temporary_storage,
                                     size_t&                     storage_size,
                                     InputIterator               input,
@@ -815,23 +859,21 @@ inline hipError_t partition_two_way(void*                       temporary_storag
 /// // output_count: 4
 /// \endcode
 /// \endparblock
-template<
-    class Config = default_config,
-    class InputIterator,
-    class FlagIterator,
-    class OutputIterator,
-    class SelectedCountOutputIterator,
-    bool  UsingOrderedBlockId = false>
-inline
-hipError_t partition(void * temporary_storage,
-                     size_t& storage_size,
-                     InputIterator input,
-                     FlagIterator flags,
-                     OutputIterator output,
-                     SelectedCountOutputIterator selected_count_output,
-                     const size_t size,
-                     const hipStream_t stream = 0,
-                     const bool debug_synchronous = false)
+template<class Config = default_config,
+         class InputIterator,
+         class FlagIterator,
+         class OutputIterator,
+         class SelectedCountOutputIterator,
+         bool UsingOrderedBlockId = false>
+inline hipError_t partition(void*                       temporary_storage,
+                            size_t&                     storage_size,
+                            InputIterator               input,
+                            FlagIterator                flags,
+                            OutputIterator              output,
+                            SelectedCountOutputIterator selected_count_output,
+                            const size_t                size,
+                            const hipStream_t           stream            = 0,
+                            const bool                  debug_synchronous = false)
 {
     using unary_predicate_type = ::rocprim::empty_type; // dummy
     using inequality_op_type   = ::rocprim::empty_type; // dummy
@@ -914,7 +956,7 @@ hipError_t partition(void * temporary_storage,
 /// #include <rocprim/rocprim.hpp>///
 ///
 /// auto predicate =
-///     [] __device__ (int a) -> bool
+///     [] (int a) -> bool
 ///     {
 ///         return (a%2) == 0;
 ///     };
@@ -951,23 +993,21 @@ hipError_t partition(void * temporary_storage,
 /// // output_count: 4
 /// \endcode
 /// \endparblock
-template<
-    class Config = default_config,
-    class InputIterator,
-    class OutputIterator,
-    class SelectedCountOutputIterator,
-    class UnaryPredicate,
-    bool  UsingOrderedBlockId = false>
-inline
-hipError_t partition(void * temporary_storage,
-                     size_t& storage_size,
-                     InputIterator input,
-                     OutputIterator output,
-                     SelectedCountOutputIterator selected_count_output,
-                     const size_t size,
-                     UnaryPredicate predicate,
-                     const hipStream_t stream = 0,
-                     const bool debug_synchronous = false)
+template<class Config = default_config,
+         class InputIterator,
+         class OutputIterator,
+         class SelectedCountOutputIterator,
+         class UnaryPredicate,
+         bool UsingOrderedBlockId = false>
+inline hipError_t partition(void*                       temporary_storage,
+                            size_t&                     storage_size,
+                            InputIterator               input,
+                            OutputIterator              output,
+                            SelectedCountOutputIterator selected_count_output,
+                            const size_t                size,
+                            UnaryPredicate              predicate,
+                            const hipStream_t           stream            = 0,
+                            const bool                  debug_synchronous = false)
 {
     using flag_type          = ::rocprim::empty_type; //dummy
     using inequality_op_type = ::rocprim::empty_type; //dummy
@@ -1078,12 +1118,12 @@ hipError_t partition(void * temporary_storage,
 /// #include <rocprim/rocprim.hpp>
 ///
 /// auto first_predicate =
-///     [] __device__ (int a) -> bool
+///     [] (int a) -> bool
 ///     {
 ///         return (a%2) == 0;
 ///     };
 /// auto second_predicate =
-///     [] __device__ (int a) -> bool
+///     [] (int a) -> bool
 ///     {
 ///         return (a%3) == 0;
 ///     };
@@ -1129,46 +1169,42 @@ hipError_t partition(void * temporary_storage,
 /// // output_count:       [4, 1]
 /// \endcode
 /// \endparblock
-template <
-    class Config = default_config,
-    typename InputIterator,
-    typename FirstOutputIterator,
-    typename SecondOutputIterator,
-    typename UnselectedOutputIterator,
-    typename SelectedCountOutputIterator,
-    typename FirstUnaryPredicate,
-    typename SecondUnaryPredicate,
-    bool     UsingOrderedBlockId = false>
-inline
-hipError_t partition_three_way(void * temporary_storage,
-                               size_t& storage_size,
-                               InputIterator input,
-                               FirstOutputIterator output_first_part,
-                               SecondOutputIterator output_second_part,
-                               UnselectedOutputIterator output_unselected,
-                               SelectedCountOutputIterator selected_count_output,
-                               const size_t size,
-                               FirstUnaryPredicate select_first_part_op,
-                               SecondUnaryPredicate select_second_part_op,
-                               const hipStream_t stream = 0,
-                               const bool debug_synchronous = false)
+template<class Config = default_config,
+         typename InputIterator,
+         typename FirstOutputIterator,
+         typename SecondOutputIterator,
+         typename UnselectedOutputIterator,
+         typename SelectedCountOutputIterator,
+         typename FirstUnaryPredicate,
+         typename SecondUnaryPredicate,
+         bool UsingOrderedBlockId = false>
+inline hipError_t partition_three_way(void*                       temporary_storage,
+                                      size_t&                     storage_size,
+                                      InputIterator               input,
+                                      FirstOutputIterator         output_first_part,
+                                      SecondOutputIterator        output_second_part,
+                                      UnselectedOutputIterator    output_unselected,
+                                      SelectedCountOutputIterator selected_count_output,
+                                      const size_t                size,
+                                      FirstUnaryPredicate         select_first_part_op,
+                                      SecondUnaryPredicate        select_second_part_op,
+                                      const hipStream_t           stream            = 0,
+                                      const bool                  debug_synchronous = false)
 {
     // Dummy flag type
-    using flag_type = ::rocprim::empty_type;
-    flag_type * flags = nullptr;
+    using flag_type  = ::rocprim::empty_type;
+    flag_type* flags = nullptr;
     // Dummy inequality operation
     using inequality_op_type = ::rocprim::empty_type;
-    using offset_type = uint2;
-    using output_key_iterator_tuple = tuple<
-        FirstOutputIterator,
-        SecondOutputIterator,
-        UnselectedOutputIterator>;
+    using offset_type        = uint2;
+    using output_key_iterator_tuple
+        = tuple<FirstOutputIterator, SecondOutputIterator, UnselectedOutputIterator>;
     using output_value_iterator_tuple
         = tuple<::rocprim::empty_type*, ::rocprim::empty_type*, ::rocprim::empty_type*>;
-    rocprim::empty_type* const no_input_values = nullptr; // key only
-    const output_value_iterator_tuple no_output_values {nullptr, nullptr, nullptr}; // key only
+    rocprim::empty_type* const        no_input_values = nullptr; // key only
+    const output_value_iterator_tuple no_output_values{nullptr, nullptr, nullptr}; // key only
 
-    output_key_iterator_tuple output{ output_first_part, output_second_part, output_unselected };
+    output_key_iterator_tuple output{output_first_part, output_second_part, output_unselected};
 
     return detail::partition_impl<detail::partition_subalgo::partition_three_way,
                                   UsingOrderedBlockId,
