@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2019-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2019-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -22,12 +22,30 @@
 
 #include "../common_test_header.hpp"
 
-// required rocprim headers
-#include <rocprim/functional.hpp>
-#include <rocprim/device/device_binary_search.hpp>
+#include "../../common/utils_custom_type.hpp"
 
 // required test headers
-#include "test_utils_types.hpp"
+#include "test_utils.hpp"
+#include "test_utils_assertions.hpp"
+#include "test_utils_data_generation.hpp"
+#include "test_utils_hipgraphs.hpp"
+
+// required common headers
+#include "../../common/utils_device_ptr.hpp"
+
+// required rocprim headers
+#include <rocprim/device/config_types.hpp>
+#include <rocprim/device/detail/device_config_helper.hpp>
+#include <rocprim/device/device_binary_search.hpp>
+#include <rocprim/functional.hpp>
+#include <rocprim/types.hpp>
+
+#include <algorithm>
+#include <cmath>
+#include <cstddef>
+#include <stdint.h>
+#include <type_traits>
+#include <vector>
 
 template<class Haystack,
          class Needle,
@@ -51,13 +69,13 @@ public:
     using params = Params;
 };
 
-using custom_int2 = test_utils::custom_test_type<int>;
-using custom_double2 = test_utils::custom_test_type<double>;
+using custom_int2    = common::custom_type<int, int, true>;
+using custom_double2 = common::custom_type<double, double, true>;
 
 struct use_custom_config
 {};
 
-typedef ::testing::Types<
+using Params = ::testing::Types<
     params<int, int>,
     params<unsigned long long, unsigned long long, size_t, rocprim::greater<unsigned long long>>,
     params<float, double, unsigned int, rocprim::greater<double>>,
@@ -72,8 +90,7 @@ typedef ::testing::Types<
            use_custom_config>,
     params<custom_int2, custom_int2>,
     params<custom_double2, custom_double2, unsigned int, rocprim::greater<custom_double2>>,
-    params<int, int, size_t, rocprim::less<>, rocprim::default_config, true>>
-    Params;
+    params<int, int, size_t, rocprim::less<>, rocprim::default_config, true>>;
 
 TYPED_TEST_SUITE(RocprimDeviceBinarySearch, Params);
 
@@ -103,7 +120,7 @@ TYPED_TEST(RocprimDeviceBinarySearch, LowerBound)
 
     compare_op_type compare_op;
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -117,36 +134,23 @@ TYPED_TEST(RocprimDeviceBinarySearch, LowerBound)
             const size_t d = haystack_size / 100;
 
             // Generate data
-            std::vector<haystack_type> haystack = test_utils::get_random_data<haystack_type>(
-                haystack_size, 0, haystack_size + 2 * d, seed_value
-            );
+            std::vector<haystack_type> haystack
+                = test_utils::get_random_data_wrapped<haystack_type>(haystack_size,
+                                                                     0,
+                                                                     haystack_size + 2 * d,
+                                                                     seed_value);
             std::sort(haystack.begin(), haystack.end(), compare_op);
 
             // Use a narrower range for needles for checking out-of-haystack cases
-            std::vector<needle_type> needles = test_utils::get_random_data<needle_type>(
-                needles_size, d, haystack_size + d, seed_value
-            );
+            std::vector<needle_type> needles
+                = test_utils::get_random_data_wrapped<needle_type>(needles_size,
+                                                                   d,
+                                                                   haystack_size + d,
+                                                                   seed_value);
 
-            haystack_type * d_haystack;
-            needle_type * d_needles;
-            output_type * d_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_haystack, haystack_size * sizeof(haystack_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_needles, needles_size * sizeof(needle_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, needles_size * sizeof(output_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_haystack, haystack.data(),
-                    haystack_size * sizeof(haystack_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_needles, needles.data(),
-                    needles_size * sizeof(needle_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<haystack_type> d_haystack(haystack);
+            common::device_ptr<needle_type>   d_needles(needles);
+            common::device_ptr<output_type>   d_output(needles_size);
 
             // Calculate expected results on host
             std::vector<output_type> expected(needles_size);
@@ -157,64 +161,50 @@ TYPED_TEST(RocprimDeviceBinarySearch, LowerBound)
                     haystack.begin();
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::params::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
-            void * d_temporary_storage = nullptr;
             size_t temporary_storage_bytes;
-            HIP_CHECK(rocprim::lower_bound<config>(d_temporary_storage,
+            HIP_CHECK(rocprim::lower_bound<config>(nullptr,
                                                    temporary_storage_bytes,
-                                                   d_haystack,
-                                                   d_needles,
-                                                   d_output,
+                                                   d_haystack.get(),
+                                                   d_needles.get(),
+                                                   d_output.get(),
                                                    haystack_size,
                                                    needles_size,
                                                    compare_op,
                                                    stream,
                                                    debug_synchronous));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-
             ASSERT_GT(temporary_storage_bytes, 0);
 
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
+            common::device_ptr<void> d_temporary_storage(temporary_storage_bytes);
 
-            if (TestFixture::params::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
-            HIP_CHECK(rocprim::lower_bound<config>(d_temporary_storage,
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
+            HIP_CHECK(rocprim::lower_bound<config>(d_temporary_storage.get(),
                                                    temporary_storage_bytes,
-                                                   d_haystack,
-                                                   d_needles,
-                                                   d_output,
+                                                   d_haystack.get(),
+                                                   d_needles.get(),
+                                                   d_output.get(),
                                                    haystack_size,
                                                    needles_size,
                                                    compare_op,
                                                    stream,
                                                    debug_synchronous));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
-            std::vector<output_type> output(needles_size);
-            HIP_CHECK(
-                hipMemcpy(
-                    output.data(), d_output,
-                    needles_size * sizeof(output_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
 
-            HIP_CHECK(hipFree(d_temporary_storage));
-            HIP_CHECK(hipFree(d_haystack));
-            HIP_CHECK(hipFree(d_needles));
-            HIP_CHECK(hipFree(d_output));
+            const auto output = d_output.load();
 
-            if (TestFixture::params::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
 
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, expected));
         }
@@ -250,7 +240,7 @@ TYPED_TEST(RocprimDeviceBinarySearch, UpperBound)
 
     compare_op_type compare_op;
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         seed_type seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -263,36 +253,23 @@ TYPED_TEST(RocprimDeviceBinarySearch, UpperBound)
             const size_t d = haystack_size / 100;
 
             // Generate data
-            std::vector<haystack_type> haystack = test_utils::get_random_data<haystack_type>(
-                haystack_size, 0, haystack_size + 2 * d, seed_value
-            );
+            std::vector<haystack_type> haystack
+                = test_utils::get_random_data_wrapped<haystack_type>(haystack_size,
+                                                                     0,
+                                                                     haystack_size + 2 * d,
+                                                                     seed_value);
             std::sort(haystack.begin(), haystack.end(), compare_op);
 
             // Use a narrower range for needles for checking out-of-haystack cases
-            std::vector<needle_type> needles = test_utils::get_random_data<needle_type>(
-                needles_size, d, haystack_size + d, seed_value
-            );
+            std::vector<needle_type> needles
+                = test_utils::get_random_data_wrapped<needle_type>(needles_size,
+                                                                   d,
+                                                                   haystack_size + d,
+                                                                   seed_value);
 
-            haystack_type * d_haystack;
-            needle_type * d_needles;
-            output_type * d_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_haystack, haystack_size * sizeof(haystack_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_needles, needles_size * sizeof(needle_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, needles_size * sizeof(output_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_haystack, haystack.data(),
-                    haystack_size * sizeof(haystack_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_needles, needles.data(),
-                    needles_size * sizeof(needle_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<haystack_type> d_haystack(haystack);
+            common::device_ptr<needle_type>   d_needles(needles);
+            common::device_ptr<output_type>   d_output(needles_size);
 
             // Calculate expected results on host
             std::vector<output_type> expected(needles_size);
@@ -303,70 +280,59 @@ TYPED_TEST(RocprimDeviceBinarySearch, UpperBound)
                     haystack.begin();
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::params::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
-            void * d_temporary_storage = nullptr;
             size_t temporary_storage_bytes;
-            HIP_CHECK(rocprim::upper_bound<config>(d_temporary_storage,
+            HIP_CHECK(rocprim::upper_bound<config>(nullptr,
                                                    temporary_storage_bytes,
-                                                   d_haystack,
-                                                   d_needles,
-                                                   d_output,
+                                                   d_haystack.get(),
+                                                   d_needles.get(),
+                                                   d_output.get(),
                                                    haystack_size,
                                                    needles_size,
                                                    compare_op,
                                                    stream,
                                                    debug_synchronous));
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
+
             ASSERT_GT(temporary_storage_bytes, 0);
 
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
+            common::device_ptr<void> d_temporary_storage(temporary_storage_bytes);
 
-            if (TestFixture::params::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
-            HIP_CHECK(rocprim::upper_bound<config>(d_temporary_storage,
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
+            HIP_CHECK(rocprim::upper_bound<config>(d_temporary_storage.get(),
                                                    temporary_storage_bytes,
-                                                   d_haystack,
-                                                   d_needles,
-                                                   d_output,
+                                                   d_haystack.get(),
+                                                   d_needles.get(),
+                                                   d_output.get(),
                                                    haystack_size,
                                                    needles_size,
                                                    compare_op,
                                                    stream,
                                                    debug_synchronous));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
-            std::vector<output_type> output(needles_size);
-            HIP_CHECK(
-                hipMemcpy(
-                    output.data(), d_output,
-                    needles_size * sizeof(output_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
 
-            HIP_CHECK(hipFree(d_temporary_storage));
-            HIP_CHECK(hipFree(d_haystack));
-            HIP_CHECK(hipFree(d_needles));
-            HIP_CHECK(hipFree(d_output));
+            const auto output = d_output.load();
 
-            if (TestFixture::params::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
 
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, expected));
         }
     }
 
-    if (TestFixture::params::use_graphs)
+    if(TestFixture::params::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TYPED_TEST(RocprimDeviceBinarySearch, BinarySearch)
@@ -395,7 +361,7 @@ TYPED_TEST(RocprimDeviceBinarySearch, BinarySearch)
 
     compare_op_type compare_op;
 
-    for (size_t seed_index = 0; seed_index < random_seeds_count + seed_size; seed_index++)
+    for(size_t seed_index = 0; seed_index < number_of_runs; seed_index++)
     {
         unsigned int seed_value = seed_index < random_seeds_count  ? rand() : seeds[seed_index - random_seeds_count];
         SCOPED_TRACE(testing::Message() << "with seed = " << seed_value);
@@ -409,36 +375,23 @@ TYPED_TEST(RocprimDeviceBinarySearch, BinarySearch)
             const size_t d = haystack_size / 100;
 
             // Generate data
-            std::vector<haystack_type> haystack = test_utils::get_random_data<haystack_type>(
-                haystack_size, 0, haystack_size + 2 * d, seed_value
-            );
+            std::vector<haystack_type> haystack
+                = test_utils::get_random_data_wrapped<haystack_type>(haystack_size,
+                                                                     0,
+                                                                     haystack_size + 2 * d,
+                                                                     seed_value);
             std::sort(haystack.begin(), haystack.end(), compare_op);
 
             // Use a narrower range for needles for checking out-of-haystack cases
-            std::vector<needle_type> needles = test_utils::get_random_data<needle_type>(
-                needles_size, d, haystack_size + d, seed_value
-            );
+            std::vector<needle_type> needles
+                = test_utils::get_random_data_wrapped<needle_type>(needles_size,
+                                                                   d,
+                                                                   haystack_size + d,
+                                                                   seed_value);
 
-            haystack_type * d_haystack;
-            needle_type * d_needles;
-            output_type * d_output;
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_haystack, haystack_size * sizeof(haystack_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_needles, needles_size * sizeof(needle_type)));
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_output, needles_size * sizeof(output_type)));
-            HIP_CHECK(
-                hipMemcpy(
-                    d_haystack, haystack.data(),
-                    haystack_size * sizeof(haystack_type),
-                    hipMemcpyHostToDevice
-                )
-            );
-            HIP_CHECK(
-                hipMemcpy(
-                    d_needles, needles.data(),
-                    needles_size * sizeof(needle_type),
-                    hipMemcpyHostToDevice
-                )
-            );
+            common::device_ptr<haystack_type> d_haystack(haystack);
+            common::device_ptr<needle_type>   d_needles(needles);
+            common::device_ptr<output_type>   d_output(needles_size);
 
             // Calculate expected results on host
             std::vector<output_type> expected(needles_size);
@@ -447,69 +400,57 @@ TYPED_TEST(RocprimDeviceBinarySearch, BinarySearch)
                 expected[i] = std::binary_search(haystack.begin(), haystack.end(), needles[i], compare_op);
             }
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::params::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
-            void * d_temporary_storage = nullptr;
             size_t temporary_storage_bytes;
-            HIP_CHECK(rocprim::binary_search<config>(d_temporary_storage,
+            HIP_CHECK(rocprim::binary_search<config>(nullptr,
                                                      temporary_storage_bytes,
-                                                     d_haystack,
-                                                     d_needles,
-                                                     d_output,
+                                                     d_haystack.get(),
+                                                     d_needles.get(),
+                                                     d_output.get(),
                                                      haystack_size,
                                                      needles_size,
                                                      compare_op,
                                                      stream,
                                                      debug_synchronous));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             ASSERT_GT(temporary_storage_bytes, 0);
 
-            HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
+            common::device_ptr<void> d_temporary_storage(temporary_storage_bytes);
 
-            if (TestFixture::params::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
-            HIP_CHECK(rocprim::binary_search<config>(d_temporary_storage,
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.startStreamCapture(stream);
+            }
+
+            HIP_CHECK(rocprim::binary_search<config>(d_temporary_storage.get(),
                                                      temporary_storage_bytes,
-                                                     d_haystack,
-                                                     d_needles,
-                                                     d_output,
+                                                     d_haystack.get(),
+                                                     d_needles.get(),
+                                                     d_output.get(),
                                                      haystack_size,
                                                      needles_size,
                                                      compare_op,
                                                      stream,
                                                      debug_synchronous));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
-            std::vector<output_type> output(needles_size);
-            HIP_CHECK(
-                hipMemcpy(
-                    output.data(), d_output,
-                    needles_size * sizeof(output_type),
-                    hipMemcpyDeviceToHost
-                )
-            );
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
 
-            HIP_CHECK(hipFree(d_temporary_storage));
-            HIP_CHECK(hipFree(d_haystack));
-            HIP_CHECK(hipFree(d_needles));
-            HIP_CHECK(hipFree(d_output));
+            const auto output = d_output.load();
 
-            if (TestFixture::params::use_graphs)
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.cleanupGraphHelper();
+            }
 
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(output, expected));
         }
     }
 
-    if (TestFixture::params::use_graphs)
+    if(TestFixture::params::use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
