@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2022-2024 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2022-2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -21,23 +21,35 @@
 // SOFTWARE.
 
 #include "../common_test_header.hpp"
-#include "rocprim/types.hpp"
+
+#include "../../common/utils.hpp"
+#include "../../common/utils_device_ptr.hpp"
+
 #include "test_utils.hpp"
+#include "test_utils_data_generation.hpp"
 
+#include <rocprim/config.hpp>
+#include <rocprim/type_traits.hpp>
+#include <rocprim/types.hpp>
 #include <rocprim/warp/warp_load.hpp>
-#include <type_traits>
 
-template<
-    class T,
-    unsigned int ItemsPerThread,
-    unsigned int WarpSize,
-    ::rocprim::warp_load_method Method
->
+#include <algorithm>
+#include <cstddef>
+#include <iterator>
+#include <numeric>
+#include <stdint.h>
+#include <type_traits>
+#include <vector>
+
+template<class T,
+         unsigned int                ItemsPerThread,
+         unsigned int                VirtualWaveSize,
+         ::rocprim::warp_load_method Method>
 struct Params
 {
     using type = T;
     static constexpr unsigned int items_per_thread = ItemsPerThread;
-    static constexpr unsigned int warp_size = WarpSize;
+    static constexpr unsigned int                warp_size        = VirtualWaveSize;
     static constexpr ::rocprim::warp_load_method method = Method;
 };
 
@@ -111,8 +123,9 @@ template<unsigned int                BlockSize,
          unsigned int                LogicalWarpSize,
          ::rocprim::warp_load_method Method,
          class T>
-__device__ auto warp_load_test(T* d_input, T* d_output)
-    -> std::enable_if_t<test_utils::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
+__device__
+auto warp_load_test(T* d_input, T* d_output)
+    -> std::enable_if_t<common::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
 {
     static_assert(BlockSize % LogicalWarpSize == 0,
                   "LogicalWarpSize must be a divisor of BlockSize");
@@ -137,8 +150,9 @@ template<unsigned int                BlockSize,
          unsigned int                LogicalWarpSize,
          ::rocprim::warp_load_method Method,
          class T>
-__device__ auto warp_load_test(T* /*d_input*/, T* /*d_output*/)
-    -> std::enable_if_t<!test_utils::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
+__device__
+auto warp_load_test(T* /*d_input*/, T* /*d_output*/)
+    -> std::enable_if_t<!common::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
 {}
 
 template<unsigned int                BlockSize,
@@ -156,8 +170,9 @@ template<unsigned int                BlockSize,
          unsigned int                LogicalWarpSize,
          ::rocprim::warp_load_method Method,
          class T>
-__device__ auto warp_load_guarded_test(T* d_input, T* d_output, int valid_items, T oob_default)
-    -> std::enable_if_t<test_utils::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
+__device__
+auto warp_load_guarded_test(T* d_input, T* d_output, int valid_items, T oob_default)
+    -> std::enable_if_t<common::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
 {
     static_assert(BlockSize % LogicalWarpSize == 0,
                   "LogicalWarpSize must be a divisor of BlockSize");
@@ -188,9 +203,9 @@ template<unsigned int                BlockSize,
          unsigned int                LogicalWarpSize,
          ::rocprim::warp_load_method Method,
          class T>
-__device__ auto
-    warp_load_guarded_test(T* /*d_input*/, T* /*d_output*/, int /*valid_items*/, T /*oob_default*/)
-        -> std::enable_if_t<!test_utils::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
+__device__
+auto warp_load_guarded_test(T* /*d_input*/, T* /*d_output*/, int /*valid_items*/, T /*oob_default*/)
+    -> std::enable_if_t<!common::device_test_enabled_for_warp_size_v<LogicalWarpSize>>
 {}
 
 template<unsigned int                BlockSize,
@@ -244,29 +259,22 @@ TYPED_TEST(WarpLoadTest, WarpLoad)
     std::vector<T> input(items_count);
     std::iota(input.begin(), input.end(), static_cast<T>(0));
 
-    T* d_input{};
-    HIP_CHECK(hipMalloc(&d_input, items_count * sizeof(T)));
-    HIP_CHECK(hipMemcpy(d_input, input.data(), items_count * sizeof(T), hipMemcpyHostToDevice));
-    T* d_output{};
-    HIP_CHECK(hipMalloc(&d_output, items_count * sizeof(T)));
+    common::device_ptr<T> d_input(input);
+    common::device_ptr<T> d_output(items_count);
 
     warp_load_kernel<block_size, items_per_thread, warp_size, method>
-        <<<dim3(1), dim3(block_size), 0, 0>>>(d_input, d_output);
+        <<<dim3(1), dim3(block_size), 0, 0>>>(d_input.get(), d_output.get());
     HIP_CHECK(hipGetLastError());
     HIP_CHECK(hipDeviceSynchronize());
 
-    std::vector<T> output(items_count);
-    HIP_CHECK(hipMemcpy(output.data(), d_output, items_count * sizeof(T), hipMemcpyDeviceToHost));
-
-    HIP_CHECK(hipFree(d_input));
-    HIP_CHECK(hipFree(d_output));
+    std::vector<T> output = d_output.load();
 
     auto expected = input;
     if(method == ::rocprim::warp_load_method::warp_load_striped)
     {
         expected = stripe_vector(input, warp_size, items_per_thread);
     }
-    
+
     ASSERT_EQ(expected, output);
 }
 
@@ -279,7 +287,8 @@ TYPED_TEST(WarpLoadTest, WarpLoadGuarded)
     constexpr unsigned int block_size = 1024;
     constexpr unsigned int items_count = items_per_thread * block_size;
     constexpr unsigned int valid_items = warp_size / 4;
-    constexpr T oob_default = std::numeric_limits<T>::max();
+
+    const T oob_default = rocprim::numeric_limits<T>::max();
 
     int device_id = test_common_utils::obtain_device_from_ctest();
     SKIP_IF_UNSUPPORTED_WARP_SIZE(warp_size, device_id);
@@ -287,22 +296,18 @@ TYPED_TEST(WarpLoadTest, WarpLoadGuarded)
     std::vector<T> input(items_count);
     std::iota(input.begin(), input.end(), static_cast<T>(0));
 
-    T* d_input{};
-    HIP_CHECK(hipMalloc(&d_input, items_count * sizeof(T)));
-    HIP_CHECK(hipMemcpy(d_input, input.data(), items_count * sizeof(T), hipMemcpyHostToDevice));
-    T* d_output{};
-    HIP_CHECK(hipMalloc(&d_output, items_count * sizeof(T)));
+    common::device_ptr<T> d_input(input);
+    common::device_ptr<T> d_output(items_count);
 
     warp_load_guarded_kernel<block_size, items_per_thread, warp_size, method>
-        <<<dim3(1), dim3(block_size), 0, 0>>>(d_input, d_output, valid_items, oob_default);
+        <<<dim3(1), dim3(block_size), 0, 0>>>(d_input.get(),
+                                              d_output.get(),
+                                              valid_items,
+                                              oob_default);
     HIP_CHECK(hipGetLastError());
     HIP_CHECK(hipDeviceSynchronize());
 
-    std::vector<T> output(items_count);
-    HIP_CHECK(hipMemcpy(output.data(), d_output, items_count * sizeof(T), hipMemcpyDeviceToHost));
-
-    HIP_CHECK(hipFree(d_input));
-    HIP_CHECK(hipFree(d_output));
+    std::vector<T> output = d_output.load();
 
     auto expected = input;
     for(size_t warp_idx = 0; warp_idx < block_size / warp_size; ++warp_idx)
@@ -311,11 +316,11 @@ TYPED_TEST(WarpLoadTest, WarpLoadGuarded)
         auto segment_end = std::next(expected.begin(), (warp_idx + 1) * warp_size * items_per_thread);
         std::fill(std::next(segment_begin, valid_items), segment_end, oob_default);
     }
-    
+
     if(method == ::rocprim::warp_load_method::warp_load_striped)
     {
         expected = stripe_vector(expected, warp_size, items_per_thread);
     }
-    
+
     ASSERT_EQ(expected, output);
 }
